# GreeseCGI.pm - Greese sample cgi script
# 2002-5 <nakajima@netstock.co.jp>
package GreeseCGI;
use Greese;
use DBI;
use CGI;
use FileHandle;
use strict;
use vars qw($VERSION);

$VERSION = 0.41;

sub main {
	my(%cfg) = @_;
	$| = 1;
	my $cgi = $cfg{cgi} || CGI->new;
	my $method = $cgi->request_method;
	if( $cfg{auth} ) {
		my $command = $cgi->param('.command');
		my @noauthcommands = 
			$cfg{auth}{noauthcommands} ? @{$cfg{auth}{noauthcommands}} : ();
		if( $method =~ /^POST$/i && $command eq 'auth' ) {
			my($orgurl, $realm, $user, $password) = 
				map {$cgi->param($_)} qw(orgurl realm user password);
			my $func = $cfg{auth}{checkpass};
			my($result, $group) = &$func($realm, $user, $password);
			if( $result eq 'OK' ) {
				my $func = $cfg{auth}{makecookie};
				my $cookie = &$func($cgi, $realm, $user, $group);
				print $cgi->redirect(-location => $orgurl, -cookie => $cookie);
				exit;
			} elsif( $result eq 'NO' ) {
				$method = 'GET';
				$cgi->param(orgurl => $orgurl);
				$cgi->param(template => $cfg{auth}{template});
				$cgi->param(realm => $realm);
				$cgi->param(checkpass => 'NO');
			} else {
				fatal($cgi, $cfg{charset}, $result);
			}
		} elsif( @noauthcommands && $method =~ /^POST$/i && 
			grep {$_ eq $command} @noauthcommands ) {
			# no auth
		} else {
			my $func = $cfg{auth}{checkcookie};
			my($result, $realm, $msg, $user, $group) = &$func($cgi, %cfg);
			if( $result eq 'OK' ) {
				$cfg{authuser} = $user;
				$cfg{authgroup} = $group;
			}
			if( $result eq 'NO' ) {
				if( $method =~ /^POST$/i ) {
					$result = 'ERROR';
					$msg = "Sorry, you are not authorized to do this action: $command";
				} else {
					$cgi->param(orgurl => $cgi->self_url);
					$cgi->param(template => $cfg{auth}{template});
					$cgi->param(realm => $realm);
				}
			}
			if( $result eq 'ERROR' ) {
				fatal($cgi, $cfg{charset}, $msg);
			}
		}
	}
	if( $method =~ /^GET$/i ) {
		GET($cgi, %cfg);
	} else {
		POST($cgi, %cfg);
	}
}

sub fatal {
	my($cgi, $charset, $msg) = @_;
	result($cgi, $charset, 'Error', $msg, 'back()');
	exit;
}

sub result {
	my($cgi, $charset, $title, $msg, $url) = @_;
	my $onclick = 
		$url eq 'close()' ? 'if(opener)opener.location.reload();window.close()' :
		$url eq 'back()' ? 'history.back()' :
		$url ? "window.location.href = '$url'" : 
		'history.back()';
	print $cgi->header(-type => 'text/html', -charset => $charset);
	print <<END;
<html><head>
<title>$title</title>
<meta http-equiv="Content-Type" content="text/html; charset=$charset">
</head>
<body bgcolor="white">
<p>$msg</p>
<form action="" method="GET">
<input type="button" value="OK" onclick="$onclick">
</form>
</body></html>
END
}

sub templateresult {
	my($cgi, $cfg, $dbh, $templatefile, %param) = @_;
	my $templateclass = $cfg->{templateclass} || 'Greese::Template::Html';
	my $tobj = $templateclass->new("$cfg->{templatedir}/$templatefile");
	my $robj = Greese::Report->new($tobj);
	$dbh ||= DBI->connect($cfg->{db}, $cfg->{user}, $cfg->{password});
	fatal($cgi, $cfg->{charset}, "cannot connect to the database '$cfg->{db}'") 
		unless $dbh;
	$dbh->trace($cfg->{dbitrace}) if $cfg->{dbitrace};
	print $cgi->header(-type => 'text/html', -charset => $cfg->{charset});
	$param{authuser} = $cfg->{authuser};
	$param{authgroup} = $cfg->{authgroup};
	$param{dbihandle} = $dbh;
	$param{cgi} = $cgi->url;
	$param{empty} = 1;
	$robj->print(%param);
}


sub GET {
	my($cgi, %cfg) = @_;
	my $selfurl = $cgi->url; 
	my %param = map {$_ => $cgi->param($_)} $cgi->param;
	my $tfile = delete $param{template};
	if( $tfile eq '.csvup' ) {
		do_getcsvup($cgi, $selfurl, $param{table}, $param{tablenum}, \%cfg);
	} elsif( $tfile eq '.csvdown' ) {
		do_getcsvdown($cgi, $selfurl, $param{table}, $param{tablenum}, \%cfg);
	} elsif( $tfile eq '.create' ) {
		do_getcreate($cgi, $selfurl, \%cfg);
	} elsif( $param{'.makehtml'} ) {
		delete $param{'.makehtml'};
		my $topleveltemplates = $cfg{topleveltemplates};
		my %already;
		print $cgi->header(-type => 'text/plain', -charset => $cfg{charset});
		print "making html files...\n";
		for my $template(@$topleveltemplates) {
			print makehtml(undef, \%already, $cgi, $selfurl, \%cfg, 
				$template, 1, %param);
		}
		print "done.\n";
	} else {
		templateresult($cgi, \%cfg, undef, $tfile, %param);
	}
}

sub makehtml1 {
	my($cgi, $cfg, $tfile, %param) = @_;
	makehtml(undef, {}, $cgi, $cgi->url, $cfg, $tfile, 0, %param);
}

sub makehtml {
	my($filename, $already, $cgi, $selfurl, $cfg, $tfile, $seeklink, %param) 
		= @_;
	my $result = '';
	my $templateclass = $cfg->{templateclass} || 'Greese::Template::Html';
	my $tobj = eval { $templateclass->new("$cfg->{templatedir}/$tfile") };
	return "ERROR in '$tfile': '$@'\n" if $@;
	my $robj = Greese::Report->new($tobj);
	my $dbh = DBI->connect($cfg->{db}, $cfg->{user}, $cfg->{password});
	fatal($cgi, $cfg->{charset}, "cannot connect to the database '$cfg->{db}'") 
		unless $dbh;
	my $filename ||= param2filename($tfile, \%param);
	$already->{$filename} = 1;
	$param{dbihandle} = $dbh;
	$param{cgi} = $selfurl;
	$param{empty} = 1;
	my $html = eval { $robj->output(%param) };
	return "ERROR in '$tfile': '$@'\n" if $@;
	my $shtml = "";
	my($host, $selfpath) = $selfurl =~ /^(https?:\/\/[^\/]+)(.+)/;
	while( $html =~ / href="(?:\Q$host\E)?\Q$selfpath\E\?(.+?)"/i ) {
		$shtml .= $`;
		$html = $';
		my $tparamstr = $1;
		my %tparam;
		for(split /&/, $tparamstr) {
			$_ = url_unescape($_);
			if( /=/ ) {
				$tparam{$`} = $';
			} else {
				$tparam{$_} = '';
			}
		}
		my $ttfile = delete $tparam{template};
		if( !$seeklink || grep {$ttfile eq $_} @{$cfg->{nomakehtml}} ) {
			$shtml .= " href=\"$selfurl?$tparamstr\"";
		} else {
			my $tfilename = param2filename($ttfile, \%tparam);
			my $turl = url_escape($tfilename);
			$shtml .= " href=\"$cfg->{htmlurl}/$turl\"";
			$result .= makehtml($tfilename, $already, $cgi, $selfurl, $cfg, 
				$ttfile, $seeklink, %tparam) unless $already->{$tfilename};
		}
	}
	$shtml .= $html;
	my $filepath = "$cfg->{htmldir}/$filename";
	my $fh = new FileHandle ">$filepath.new";
	if( $fh ) {
		print $fh $shtml;
		close $fh;
		if( rename "$filepath.new", $filepath ) {
			$result .= "$filepath\n";
		} else {
			$result .= "ERROR: cannot mv $filepath.new $filepath\n";
		}
	} else {
		$result .= "ERROR: cannot open $filepath.new\n";
	}
	$result;
}

sub param2filename {
	my($template, $param) = @_;
	my $body = $template;
	my $ext = '';
	if( $body =~ s/(\.[^\.]+)$// ) {
		$ext = $1;
	}
	my $filename = $body;
	for my $key(sort keys %$param) {
		my $value = $param->{$key};
		$filename .= '_' if $filename ne '';
		$filename .= $key . '_' . $value;
	}
	$filename .= $ext;
	$filename;
}

sub do_getcsvup {
	my($cgi, $selfurl, $table, $tablenum, $cfg) = @_;
	my $charset = $cfg->{charset};
	print $cgi->header(-type => 'text/html', -charset => $charset);
	print <<END;
<html><head>
<title>CSV data file upload</title>
<meta http-equiv="Content-Type" content="text/html; charset=$charset">
</head>
<body bgcolor="white">
<h3>CSV data file upload</h3>
<form action="$selfurl" method="post" enctype="multipart/form-data">
<input type="hidden" name=".command" value="csv">
<input type="hidden" name="tablenum" value="$tablenum">
<table>
  <tr>
    <td>table name</td>
    <td>$table<input type="@{[$table ? 'hidden' : 'text']}" name=".csvtable" size="40" value="$table"></td>
  </tr>
  <tr>
    <td>csv file</td>
    <td><input type="file" name=".csvfile" size="80"></td>
  </tr>
  <tr>
    <td></td>
    <td><input type="checkbox" name=".cleartable" value="1"> clear the table before inserting data</td>
  </tr>
</table>
<input type="submit">
<input type="reset">
</form>
</body></html>
END
}

sub do_getcsvdown {
	my($cgi, $selfurl, $table, $tablenum, $cfg) = @_;
	my $charset = $cfg->{charset};
	#print $cgi->header(-type => 'text/plain', -charset => $charset);
	print $cgi->header(-type => 'application/octet-stream', -charset => $charset);
	my @fields = grep {!/>/} fields($cfg, $table);
	my %mark;
	grep {if(s/^([*@#&]+)//){$mark{$_} = $1}} @fields;
	my $cols = join(",", map {quotename($_)} @fields);
	print join(",", @fields), "\n";
	my $dbh = DBI->connect($cfg->{db}, $cfg->{user}, $cfg->{password});
	for my $row( @{$dbh->selectall_arrayref("select $cols from \"$table\"")} ) {
		print join(",", map {'"'.escape_quote($_).'"'} @$row),"\n";
	}
}

sub do_getcreate {
	my($cgi, $selfurl, $cfg) = @_;
	my $charset = $cfg->{charset};
	print $cgi->header(-type => 'text/html', -charset => $charset);
	print <<END;
<html><head>
<title>create tables</title>
<meta http-equiv="Content-Type" content="text/html; charset=$charset">
</head>
<body bgcolor="white">
<h3>Table definition data file upload</h3>
<form action="$selfurl" method="post" enctype="multipart/form-data">
<input type="hidden" name=".command" value="create">
<table>
  <tr>
    <td>password</td>
    <td><input type="password" name=".password" size="20"></td>
  </tr>
  <tr>
    <td>data file</td>
    <td><input type="file" name=".datafile" size="80"></td>
  </tr>
  <tr>
    <td></td>
    <td><input type="checkbox" name=".templateonly" value="1">make only template files</td>
  </tr>
  <tr>
    <td></td>
    <td><input type="checkbox" name=".notemplate" value="1">not make template files</td>
  </tr>
</table>
<input type="submit">
<input type="reset">
</form>
</body></html>
END
}

sub POST {
	my($cgi, %cfg) = @_;
	if( $cfg{custom_command} ) {
		my $command = $cgi->param('.command');
		if( $cfg{custom_command}{$command} ) {
			my $func = $cfg{custom_command}{$command};
			my($title, $msg, $url) = &$func($cgi, \%cfg);
			result($cgi, $cfg{charset}, $title, $msg, $url);
			return;
		}
	}
	my %para;
	for my $para($cgi->param) {
		if( $para =~ /^\./ ) {
			$cfg{$para} = $cgi->param($para);
		} else {
			my @values = $cgi->param($para);
			$para{$para} = @values > 1 ? \@values : $values[0];
		}
	}
	my $command = $cfg{'.command'};
	my $fromwhere = url_unescape($cfg{'.fromwhere'});
	if( $command =~ /^update$/i ) {
		do_update($cgi, \%cfg, $fromwhere, \%para);
	} elsif( $command =~ /^insert$/i ) {
		do_insert($cgi, \%cfg, $fromwhere, \%para);
	} elsif( $command =~ /^delete$/i ) {
		do_delete($cgi, \%cfg, $fromwhere, \%para);
	} elsif( $command =~ /^create$/i ) {
		do_create($cgi, \%cfg, \%para);
	} elsif( $command =~ /^csv$/i ) {
		do_csv($cgi, \%cfg, \%para);
	} elsif( $command =~ /^select$/i ) {
		do_select($cgi, \%cfg, $fromwhere, \%para);
	} else {
		fatal($cgi, $cfg{charset}, 'unknown command: '.$command);
	}
}

sub do_update {
	my($cgi, $cfg, $fromwhere, $para) = @_;
	my($dbh, $table, $where) = do_prepare('update', $cgi, $cfg, $fromwhere);
	$dbh->begin_work;
	my $sep = $cfg->{multivalue_separator} || ',';
	my $rtable = dequotename($table);
	my %listtable;
	my $fieldmarks = fieldmarks($cfg, $rtable);
	for my $tf(keys %$fieldmarks) {
		if( $tf =~ />/ ) {
			my($field, $ltname) = ($`, $');
			$listtable{$field} = $ltname;
			$para->{$field} = '' unless exists $para->{$field};
		}
	}
	my(%values, %needquote, $primaryvalue);
	for my $field(keys %$para) {
		my $value = $para->{$field};
		my $mark;
		if( $field =~ s/^([*@#&]+)// ) {
			$mark = $1;
		}
		if( ref($value) eq 'ARRAY' && !$listtable{$field} ) {
			$value = join($sep, @$value);
		}
		my $needquote = 1;
		my $notset = 0;
		if( $mark =~ /\*/ ) { # primary 
			$primaryvalue = $value;
		}
		if( $mark =~ /#/ ) { # number field
			$needquote = 0;
		} elsif( $mark =~ /\@/ ) { # current date field
			$value = now();
		} elsif( $mark =~ /&/ ) { # file field
			my $filename = $value;
			if( $filename ) {
				$value = savefile($cgi, $cfg, $filename);
				deletefile($cfg, fieldvalue($cfg, $field, $table, $where));
			} elsif( $cfg->{".clear&$field"} ) {
				deletefile($cfg, fieldvalue($cfg, $field, $table, $where));
			} else {
				$notset = 1;
			}
		}
		$values{$field} = $value unless $notset;
		$needquote{$field} = $needquote;
	}
	if( $cfg->{beforeupdate}{$rtable} ) {
		my $callback = $cfg->{beforeupdate}{$rtable};
		my($result, $newvalue) = 
			&$callback($cgi, 'update', $cfg, $dbh, $rtable, \%values, $where);
		if( $result eq 'OK' ) {
			%values = %$newvalue;
		} else {
			$dbh->rollback;
			fatal($cgi, $cfg->{charset}, $result);
		}
	}
	my @set;
	for my $field(keys %values) {
		next if $listtable{$field};
		my $value = $values{$field};
		$value = 'null' if !$needquote{$field} && $value eq '';
		push @set, quotename($field)." = ".
			($needquote{$field} ? $dbh->quote($value) : $value);
	}
	my($sql, @sqls);
	$sql = "update $table set ";
	$sql .= join(',', @set) . " " . $where;
	push @sqls, $sql;
	for my $field(keys %values) {
		if( $listtable{$field} ) {
			my($ltf1name, $ltf2name) = fields($cfg, $listtable{$field});
			my($ltf1mark, $ltf2mark, $quote1, $quote2);
			if( $ltf1name =~ s/^([*@#&]+)// ) {
				$ltf1mark = $1;
				$quote1 = "'" unless $ltf1mark =~ /#/;
			}
			if( $ltf2name =~ s/^([*@#&]+)// ) {
				$ltf2mark = $1;
				$quote2 = "'" unless $ltf1mark =~ /#/;
			}
			my @values = ref($values{$field}) eq 'ARRAY' ? 
				@{$values{$field}} : ($values{$field});
			@values = grep {$_ ne ''} @values;
			my @current = map {$_->[0]} @{$dbh->selectall_arrayref("select \"$ltf2name\" from \"$listtable{$field}\" where \"$ltf1name\" = $quote1$primaryvalue$quote1") || []};
			for my $insv(arraysub(\@values, \@current)) {
				$sql = "insert into \"$listtable{$field}\" (\"$ltf1name\", \"$ltf2name\") values ($quote1$primaryvalue$quote1, $quote2$insv$quote2)";
				push @sqls, $sql;
			}
			for my $delv(arraysub(\@current, \@values)) {
				$sql = "delete from \"$listtable{$field}\" where \"$ltf1name\" = $quote1$primaryvalue$quote1 and \"$ltf2name\" = $quote2$delv$quote2";
				push @sqls, $sql;
			}
		}
	}
	my $r = do_exec($cgi, $dbh, $cfg, $where, @sqls);
	if( $r ) {
		$dbh->commit;
	} else {
		$dbh->rollback;
	}
	if( $r && $cfg->{afterupdate}{$rtable} ) {
		my $callback = $cfg->{afterupdate}{$rtable};
		&$callback($cgi, 'update', $cfg, $dbh, $rtable, \%values, $where);
	}
}

sub arraysub {
	my($a, $b) = @_;
	my %tmp;
	for(@$a) {
		$tmp{$_} = 1;
	}
	for(@$b) {
		delete $tmp{$_};
	}
	keys %tmp;
}

sub fieldvalue {
	my($cfg, $field, $table, $where) = @_;
	my $dbh = DBI->connect($cfg->{db}, $cfg->{user}, 
		$cfg->{password}, {AutoCommit => 1});
	my($value) = $dbh->selectrow_array("select ".
		quotename($field)." from $table $where");
	$dbh->disconnect;
	$value;
}

sub do_insert {
	my($cgi, $cfg, $fromwhere, $para) = @_;
	my($dbh, $table, $where) = do_prepare('insert', $cgi, $cfg, $fromwhere);
	$dbh->begin_work;
	my $sep = $cfg->{multivalue_separator} || ',';
	my $rtable = dequotename($table);
	my %listtable;
	my $fieldmarks = fieldmarks($cfg, $rtable);
	for my $tf(keys %$fieldmarks) {
		if( $tf =~ />/ ) {
			my($field, $ltname) = ($`, $');
			$listtable{$field} = $ltname;
		}
	}
	my(%values, %needquote, $primaryfield, $primaryvalue);
	for my $field(keys %$para) {
		my $value = $para->{$field};
		my $mark;
		if( $field =~ s/^([*@#&]+)// ) {
			$mark = $1;
		}
		if( ref($value) eq 'ARRAY' && !$listtable{$field} ) {
			$value = join($sep, @$value);
		}
		my $needquote = 1;
		if( $mark =~ /##/ ) { # auto increment number field
			$needquote = 0;
			if( $cfg->{locktablesql} ) {
				$dbh->do($cfg->{locktablesql});
			}
			#$value = "coalesce((select max(".quotename($field).") from $table), 0) + 1";
			($value) = $dbh->selectrow_array("select max(".quotename($field).") from $table");
			$value += 1;
		} elsif( $mark =~ /#/ ) { # number field
			$needquote = 0;
		} elsif( $mark =~ /\@/ ) { # current date field
			$value = now();
		} elsif( $mark =~ /&/ ) { # file field
			my $filename = $value;
			if( $filename ) {
				$value = savefile($cgi, $cfg, $filename);
			}
		}
		if( $mark =~ /\*/ ) { # primary 
			$primaryfield = $field;
			$primaryvalue = $value;
			$where = $needquote ? "where \"$field\" = '$value'" : 
				"where \"$field\" = $value";
		}
		$values{$field} = $value;
		$needquote{$field} = $needquote;
	}
	if( $cfg->{beforeupdate}{$rtable} ) {
		my $callback = $cfg->{beforeupdate}{$rtable};
		my($result, $newvalues) = 
			&$callback($cgi, 'insert', $cfg, $dbh, $rtable, \%values);
		if( $result eq 'OK' ) {
			%values = %$newvalues;
		} else {
			$dbh->rollback;
			fatal($cgi, $cfg->{charset}, $result);
		}
	}
	my(@fields, @values);
	for my $field(keys %values) {
		next if $listtable{$field};
		my $value = $values{$field};
		$value = 'null' if !$needquote{$field} && $value eq '';
		push @fields, quotename($field);
		push @values, ($needquote{$field} ? $dbh->quote($value) : $value);
	}
	my($sql, @sqls);
	$sql = "insert into $table ";
	$sql .= '('.join(',', @fields).')';
	$sql .= ' values ';
	$sql .= '('.join(',', @values).')';
	push @sqls, $sql;
	for my $field(keys %values) {
		if( $listtable{$field} ) {
			my($ltf1name, $ltf2name) = fields($cfg, $listtable{$field});
			my($ltf1mark, $ltf2mark, $quote1, $quote2);
			if( $ltf1name =~ s/^([*@#&]+)// ) {
				$ltf1mark = $1;
				$quote1 = "'" unless $ltf1mark =~ /#/;
			}
			if( $ltf2name =~ s/^([*@#&]+)// ) {
				$ltf2mark = $1;
				$quote2 = "'" unless $ltf1mark =~ /#/;
			}
			for my $value(ref($values{$field}) eq 'ARRAY' ? 
				@{$values{$field}} : ($values{$field})) {
				$sql = "insert into \"$listtable{$field}\" (\"$ltf1name\", \"$ltf2name\") values ($quote1$primaryvalue$quote1, $quote2$value$quote2)";
				push @sqls, $sql;
			}
		}
	}
	my $r = do_exec($cgi, $dbh, $cfg, $where, @sqls);
	if( $r ) {
		$dbh->commit;
	} else {
		$dbh->rollback;
	}
	if( $r && $cfg->{afterupdate}{$rtable} ) {
		my $callback = $cfg->{afterupdate}{$rtable};
		&$callback($cgi, 'insert', $cfg, $dbh, $rtable, \%values);
	}
}

sub savefile {
	my($cgi, $cfg, $filename) = @_;
	my $filedir = $cfg->{filedir};
	my $countfile = "$filedir/.count";
	my $filenum = nexfilenum($countfile);
	if( $filenum eq '' ) {
		fatal($cgi, $cfg->{charset}, "cannot get next number for file");
	}
	#my($ext) = $filename =~ /([^\.\/]+)$/;
	#$ext = lc $ext;
	#my $dataname = "$filenum.$ext";
	my($dir, $body, $ext) = splitpath($cfg->{charset}, $filename);
	#my $dataname = url_escape("$filenum/$body$ext");
	my $dataname = "$filenum/$body$ext";
	mkdir "$filedir/$filenum", 0775;
	my $fh = new FileHandle ">$filedir/$dataname"
		or fatal($cgi, $cfg->{charset}, 
		"cannot save file '$filedir/$dataname'");
	seek $filename, 0, 0;
	while(<$filename>) {
		print $fh $_;
	}
	close $fh;
	$dataname;
}

sub splitpath {
	my($charset, $path) = @_;
	my($dir, $body, $ext);
	if( $charset =~ /Shift_JIS/i ) {
		my @c = split //, $path;
		my $lastslashpos = -1;
		for(my $j = 0; $j < @c; $j++ ) {
			my $c = $c[$j];
			if( $c =~ /[\x81-\x9f\xe0-\xfc]/ ) {
				$j++;
			} elsif( $c =~ /[\\\/]/ ) {
				$lastslashpos = $j;
			}
		}
		if( $lastslashpos >= 0 ) {
			$dir = substr($path, 0, $lastslashpos + 1);
			$body = substr($path, $lastslashpos + 1);
		} else {
			$body = $path;
		}
		if( $body =~ /(\.[^\.]+)$/ ) {
			$body = $`;
			$ext = $1;
		}
	} else {
		if( $path =~ /^(.+[\\\/])/ ) {
			$dir = $1;
			$body = $';
		} else {
			$body = $path;
		}
		if( $body =~ /(\.[^\.]+)$/ ) {
			$body = $`;
			$ext = $1;
		}
	}
	($dir, $body, $ext);
}

sub deletefile {
	my($cfg, $filename) = @_;
	return unless $filename;
	my $filedir = $cfg->{filedir};
	my($dir) = $filename =~ /^(\d+)/;
	#unlink "$filedir/$filename";
	`rm -fr $filedir/$dir`;
}

sub nexfilenum {
	my($countfile) = @_;
	my $fh = new FileHandle (-f $countfile ? "+<$countfile" : "+>$countfile");
	return unless $fh;
	flock $fh, 2;
	seek $fh, 0, 0;
	my $count = <$fh>;
	$count++;
	seek $fh, 0, 0;
	print $fh $count;
	flock $fh, 8;
	close $fh;
	$count;
}

sub do_delete {
	my($cgi, $cfg, $fromwhere, $para) = @_;
	my($dbh, $table, $where) = do_prepare('delete', $cgi, $cfg, $fromwhere);
	my $rtable = dequotename($table);
	my $fieldmarks = fieldmarks($cfg, $rtable);
	my(@filefields, @filenames, %listtable, $primary);
	for my $field(keys %$fieldmarks) {
		if( $field =~ />/ ) {
			my $ltname;
			($field, $ltname) = ($`, $');
			$listtable{$field} = $ltname;
		}
		if( $fieldmarks->{$field} =~ /&/ ) {
			push @filefields, $field;
		}
		if( $fieldmarks->{$field} =~ /\*/ ) {
			$primary = $field;
		}
	}
	if( @filefields ) {
		@filenames = $dbh->selectrow_array("select ".
			join(",", map {quotename($_)} @filefields).
			" from $table $where");
	}
	my $oldvalue;
	if( $cfg->{beforeupdate}{$rtable} ) {
		$oldvalue = $dbh->selectrow_hashref("select * from $table $where");
		my $callback = $cfg->{beforeupdate}{$rtable};
		my($result) = 
			&$callback($cgi, 'delete', $cfg, $dbh, $rtable, $oldvalue);
		if( $result ne 'OK' ) {
			fatal($cgi, $cfg->{charset}, $result);
		}
	}
	my @sqls;
	for my $field(keys %listtable) {
		my($ltf1name, $ltf2name) = fields($cfg, $listtable{$field});
		my($ltf1mark, $quote1);
		if( $ltf1name =~ s/^([*@#&]+)// ) {
			$ltf1mark = $1;
			$quote1 = "'" unless $ltf1mark =~ /#/;
		}
		push @sqls, "delete from \"$listtable{$field}\" where \"$ltf1name\" in (select \"$primary\" from $table $where)";
	}
	push @sqls, "delete from $table $where";
	my $r = do_exec($cgi, $dbh, $cfg, $where, @sqls);
	if( $r && @filenames ) {
		for my $filename(@filenames) {
			deletefile($cfg, $filename) if $filename;
		}
	}
	if( $cfg->{afterupdate}{$rtable} ) {
		my $callback = $cfg->{afterupdate}{$rtable};
		&$callback($cgi, 'delete', $cfg, $dbh, $rtable, $oldvalue);
	}
}

sub do_create {
	my($cgi, $cfg, $para) = @_;
	my $datafile = $cfg->{'.datafile'}
		or fatal($cgi, $cfg->{charset}, 
		'create command needs .datafile parameter');
	my $templateonly = $cfg->{'.templateonly'};
	my $notemplate = $cfg->{'.notemplate'};
	my($dbh) = do_prepare('create', $cgi, $cfg);
	print $cgi->header(-type => 'text/html', -charset => $cfg->{charset});
	print <<END;
<html><header><title>create tables result</title></head>
<body><pre>
END
	my($tablefields, $tabletypes, $tabledisps, $tablelists, $tablepatterns, 
		$tablemarks, $tableprimary) = 
		makedata($cgi, $cfg, $datafile, $dbh, !$templateonly);
	showall($dbh, $tablefields) unless $templateonly;
	makedefaulttemplates($cgi, $dbh, $cfg->{charset}, $cfg->{templatedir}, 
		$cfg->{fileurl},
		$tablefields, $tabletypes, $tabledisps, $tablelists, $tablepatterns, 
		$tablemarks, $tableprimary,
		) unless $notemplate;
	print <<END;
</pre></body></html>
END
}

sub makedata {
	my($cgi, $cfg, $datafile, $dbh, $make) = @_;
	my $handle = $datafile;
	my($stat, $table, @def, @tf, %tablefields, %tabletypes, %tabledisps, 
		%tablelists, %tablepatterns, %tablemarks, %tableprimary, %index);
	my $tfh = new FileHandle ">$cfg->{templatedir}/tablefields"
		or fatal($cgi, $cfg->{charset}, 
			"cannot open $cfg->{templatedir}/tablefields");
	while(<$handle>) {
		s/\r\n/\n/;
		chomp;
		if( /^def\s+(.+)/ ) {
			$table = $1;
			$stat = 'def';
			@def = ();
			@tf = ();
			%index = ();
		} elsif( /^enddef$/ ) {
			if( $make ) {
				my $sql;
				$sql = "drop table \"$table\"";
				print "$sql\n";
				$dbh->do($sql);
				$sql = "create table \"$table\" (".join(",", @def).")";
				print "$sql\n";
				$dbh->do($sql);
				for my $index(keys %index) {
					$sql = "create index \"${table}_${index}_index\" on \"$table\" (\"$index\")";
					print "$sql\n";
					$dbh->do($sql);
				}
			}
			print $tfh "$table\t", join("\t", @tf), "\n";
		} elsif( /^sql\s+(.+)/ ) {
			my $sql = $1;
			print "$sql\n";
			$dbh->do($sql);
			print $dbh->errstr, "\n";
		} elsif( /^data\s+(.+)/ ) {
			$table = $1;
			$stat = 'data';
		} elsif( /^index\s+(\S+)\s+(.+)/ ) {
			my $table = $1;
			my @cols = split(/\s*,\s*/, $2);
			my $indexname = $table . '_' . join('_', @cols) . '_index';
			my $sql = "create index \"$indexname\" on \"$table\" (". 
				join(',', map {"\"$_\""} @cols) . ")";
			print "$sql\n";
			$dbh->do($sql);
		} elsif( $stat eq 'def' ) {
			s/^\s+//;
			my($colname, $type, $disp, $pattern) = split /\t/;
			if( $colname && $type ) {
				if( $colname =~ s/^\*// ) {
					$index{$colname}++;
				}
				push @{$tablefields{$table}}, $colname;
				if( $disp =~ s/^\+// ) {
					$tablelists{$table}{$colname} = 1
				}
				$tabledisps{$table}{$colname} = $disp || 60;
				$tablepatterns{$table}{$colname} = $pattern;
				my $mark = '';
				if( $type =~ s/^([@#&]+)// ) {
					$mark = $1;
				}
				$tabletypes{$table}{$colname} = $type;
				if( $mark eq '' && $type !~ /char|text|date|time|listtable/i ) {
					$mark = '#';
				}
				if( $type =~ /primary\s+key/i ) {
					$tableprimary{$table} = $colname;
					$tablelists{$table}{$colname} = 1;
					$mark .= '*';
				}
				$tablemarks{$table}{$colname} = $mark;
				if( $type =~ /^listtable\((.+)\)/ ) {
					my $ltspec = $1;
					my($ltname, $ltf1name, $ltf1type, $ltf2name, $ltf2type)
						= splitlisttablespec($ltspec);
					my $ltf1mark = 
						$ltf1type !~ /char|text|date|time/i ? '#' : '';
					my $ltf2mark = 
						$ltf2type !~ /char|text|date|time/i ? '#' : '';
					$tablefields{$ltname} = [$ltf1name, $ltf2name];
					$tablelists{$ltname}{$ltf1name} = 1;
					$tablelists{$ltname}{$ltf2name} = 1;
					$tabletypes{$ltname}{$ltf1name} = $ltf1type;
					$tabletypes{$ltname}{$ltf2name} = $ltf2type;
					$tablemarks{$ltname}{$ltf1name} = $ltf1mark;
					$tablemarks{$ltname}{$ltf2name} = $ltf2mark;
					$tabledisps{$ltname}{$ltf1name} = 60;
					$tabledisps{$ltname}{$ltf2name} = 60;
					print $tfh 
						"$ltname\t$ltf1mark$ltf1name\t$ltf2mark$ltf2name\n";
					if( $make ) {
						print "ERROR: illegal listtable() spec: $ltspec\n" 
							if $ltname eq '' || $ltf1name eq '' ||
								$ltf1type eq '' || $ltf2type eq '';
						my $sql;
						$sql = "drop table \"$ltname\"";
						print "$sql\n";
						$dbh->do($sql);
						$sql = "create table \"$ltname\" (\"$ltf1name\" $ltf1type, \"$ltf2name\" $ltf2type)";
						print "$sql\n";
						$dbh->do($sql);
						$sql = "create index \"${ltname}_${ltf1name}_index\" on \"$ltname\" (\"$ltf1name\")";
						print "$sql\n";
						$dbh->do($sql);
						$sql = "create index \"${ltname}_${ltf2name}_index\" on \"$ltname\" (\"$ltf2name\")";
						print "$sql\n";
						$dbh->do($sql);
					}
					push @tf, "$colname>$ltname";
				} else {
					push @def, "\"$colname\" $type";
					push @tf, "$mark$colname";
				}
			}
		} elsif( $stat eq 'data' ) {
			s/^\s+//;
			my @data = split /\t/;
			my $sql = "insert into \"$table\" values (".
				join(",", map {"'$_'"} @data).")";
			if( $make ) {
				print "$sql\n";
				$dbh->do($sql);
			}
		} 
	}
	close $tfh;
	(\%tablefields, \%tabletypes, \%tabledisps, \%tablelists,
		\%tablepatterns, \%tablemarks, \%tableprimary);
}

sub splitlisttablespec {
	my($spec) = @_;
	my($ltname, $ltf1, $ltf2, $ltselect, $ltdisp) = 
		$spec =~ /^([^,]+)\s*,\s*([^,]+)\s*,\s*([^,]+)\s*,\s*(\*\[select\(.+\)\]\*)\s*,\s*(.+)/;
	my($ltf1name, $ltf1type) = $ltf1 =~ /^(\S+)\s+(.+)/;
	my($ltf2name, $ltf2type) = $ltf2 =~ /^(\S+)\s+(.+)/;
	($ltname, $ltf1name, $ltf1type, $ltf2name, $ltf2type, $ltselect, $ltdisp);
}

sub showall {
	my($dbh, $tablefields) = @_;
	for my $table(keys %$tablefields) {
		print "$table(", join(",", @{$tablefields->{$table}}), ")\n";
		for my $row( @{$dbh->selectall_arrayref("select * from \"$table\"")} ) {
			print "  ",join(",", @$row),"\n";
		}
	}
}

sub makedefaulttemplates {
	my($cgi, $dbh, $charset, $templatedir, $fileurl, $tablefields, $tabletypes, 
		$tabledisps, $tablelists, $tablepatterns, $tablemarks, 
		$tableprimary) = @_;
	my $selfurl = $cgi->url;
	my @tnames = keys %$tablefields;
	for( my $j = 0; $j < @tnames; $j++ ) {
		my $tname = $tnames[$j];
		my @fields = @{$tablefields->{$tname}};
		my $primary = $tableprimary->{$tname};
		makelisttemplate($selfurl, $charset, $templatedir, $fileurl, $tname, $j,
			$primary, $tabletypes->{$tname}, $tabledisps->{$tname}, 
			$tablemarks->{$tname}, $tablelists->{$tname}, 
			@fields);
		if( $primary ) {
			makeviewtemplate($selfurl, $charset, $templatedir, $fileurl, 
				$tname, $j, $primary, $tabletypes->{$tname}, 
				$tabledisps->{$tname}, $tablemarks->{$tname}, 
				@fields);
			makeinserttemplate($dbh, $selfurl, $charset, $templatedir, $fileurl,
				$tname, $j, $tabletypes->{$tname}, 
				$tabledisps->{$tname}, $tablepatterns->{$tname},
				$tablemarks->{$tname}, @fields);
			makeedittemplate($dbh, $selfurl, $charset, $templatedir, $fileurl, 
				$tname, $j, $primary, $tabletypes->{$tname}, 
				$tabledisps->{$tname}, $tablepatterns->{$tname},
				$tablemarks->{$tname}, @fields);
		}
	}
	open F, ">$templatedir/default.html" or return;
	print "make template $templatedir/default.html\n";
	print F <<END;
<HTML><HEAD>
<META http-equiv="Content-Type" content="text/html; charset=$charset">
</HEAD>
<BODY BGCOLOR="white">
<UL>
END
	for( my $j = 0; $j < @tnames; $j++ ) {
		my $tname = $tnames[$j];
		my $primary = $tableprimary->{$tname};
		print F "<LI><a href=\"$selfurl?template=table$j.html\">$tname</a></LI>\n";
	}
	print F <<END;
</UL>
</BODY></HTML>
END
}

sub makelisttemplate {
	my($selfurl, $charset, $templatedir, $fileurl, $table, $number, $primary, 
		$types, $disps, $marks, $lists, @fields) = @_;
	@fields = grep { $lists->{$_} } @fields;
	my $qtable = quotename($table);
	open F, ">$templatedir/table$number.html" or return;
	print "make template $templatedir/table$number.html\n";
	print F <<END;
<HTML><HEAD>
<meta http-equiv="Content-Type" content="text/html; charset=$charset">
</HEAD>
<BODY BGCOLOR="white">
*[select(from $qtable %[param(orderby)]%)]*
<p align="right">
<a href="$selfurl?template=default.html">HOME</a>
END
	if( $primary ) {
		print F <<END;
<a href="$selfurl?template=table${number}i.html">New data</a>
END
	}
	print F <<END;
<a href="$selfurl?template=.csvup&table=*[urlescape:'$table']*&tablenum=$number">CSV upload</a>
<a href="$selfurl?template=.csvdown&table=*[urlescape:'$table']*&tablenum=$number">CSV download</a>
</p>
<h3>$qtable table *[param(orderby)]*</h3>
<table border=1>
<tr>
END
	for my $field(@fields) {
		print F "  <th><a href=\"$selfurl?template=table$number.html&orderby=order by *[urlescape:'\"$field\"']*\">$field</a></th>\n";
	}
	print F <<END;
</tr>
<!--*[repeat()]*-->
<tr>
END
	for my $field(@fields) {
		my $needquote = $marks->{$field} !~ /#/;
		my $align = $needquote ? 'left' : 'right';
		if( $field eq $primary ) {
			print F "  <td align=\"$align\"><a href=\"$selfurl?template=table${number}v.html&where=*[urlescape:'$field']*=";
			print F $needquote ? "'*[$field]*'" : "*[$field]*";
			print F "\">*[$field]*</a></td>\n";
		} elsif( $marks->{$field} =~ /&/ ) {
			print F "  <td align=\"$align\"><a href=\"$fileurl/*[urlescape:$field]*\" target=\"_blank\">*[$field]*</a></td>\n";
		} elsif( $types->{$field} =~ /^listtable\((.+)\)/ ) {
			my($ltname, $ltf1name, $ltf1type, $ltf2name, $ltf2type, 
				$ltselect, $ltdisp) = splitlisttablespec($1);
			print F "  <td align=\"$align\">$ltselect*[repeat()]*$ltdisp*[separator()]*,*[end()]**[end()]*</td>\n";
		} elsif( $disps->{$field} =~ /^\s*(\*\[select.+\]\*),([^,]+),([^,]+)\s*$/ ) {
			my($select, $selectval, $selectopt) = ($1,$2,$3);
			$selectopt = "*[$selectopt]*" unless $selectopt =~ /\*\[/;
			if( $selectopt eq "*[$selectval]*" ) {
				print F "  <td align=\"$align\">*[$field]*</td>\n";
			} else {
				my $quote = $needquote ? "'" : "";
				my $vselect = addandcond($select, 
					"\"$selectval\" = $quote%[$field]%$quote");
				print F "  <td align=\"$align\">$vselect$selectopt*[end()]*</td>\n";
			}
		} else {
			print F "  <td align=\"$align\">*[$field]*</td>\n";
		}
	}
	print F <<END;
</tr>
<!--*[end()]*-->
</table>
*[end()]*
</BODY></HTML>
END
	close F;
}

sub makeviewtemplate {
	my($selfurl, $charset, $templatedir, $fileurl, $table, $number, $primary, 
		$types, $disps, $marks, @fields) = @_;
	my $qtable = quotename($table);
	open F, ">$templatedir/table${number}v.html" or return;
	print "make template $templatedir/table${number}v.html\n";
	print F <<END;
<HTML><HEAD>
<meta http-equiv="Content-Type" content="text/html; charset=$charset">
</HEAD>
<BODY BGCOLOR="white">
*[select(from $qtable)]*
<p align="right">
<a href="$selfurl?template=default.html">HOME</a>
<a href="$selfurl?template=table$number.html">$qtable table</a>
<a href="$selfurl?template=table${number}e.html&where=*[urlescape:param(where)]*">edit this data</a>
</p>
<h3>View data in $qtable table</h3>
<table border=1>
END
	for my $field(@fields) {
		my $needquote = $marks->{$field} !~ /#/;
		#my $align = $needquote ? 'left' : 'right';
		my $align = 'left';
		if( $marks->{$field} =~ /&/ ) {
			print F "  <tr><th>$field</th><td align=\"$align\"><a href=\"$fileurl/*[urlescape:$field]*\" target=\"_blank\">*[$field]*</a></td></tr>\n";
		} elsif( $types->{$field} =~ /^listtable\((.+)\)/ ) {
			my($ltname, $ltf1name, $ltf1type, $ltf2name, $ltf2type, 
				$ltselect, $ltdisp) = splitlisttablespec($1);
			print F "  <tr><th>$field</th><td align=\"$align\">$ltselect*[repeat()]*$ltdisp*[separator()]*,*[end()]**[end()]*</td></tr>\n";
		} elsif( $disps->{$field} =~ /^\s*(\*\[select.+\]\*),([^,]+),([^,]+)\s*$/ ) {
			my($select, $selectval, $selectopt) = ($1,$2,$3);
			$selectopt = "*[$selectopt]*" unless $selectopt =~ /\*\[/;
			if( $selectopt eq "*[$selectval]*" ) {
				print F "  <tr><th>$field</th><td align=\"$align\">*[nl2br:$field]*</td></tr>\n";
			} else {
				my $quote = $needquote ? "'" : "";
				my $vselect = addandcond($select, 
					"\"$selectval\" = $quote%[$field]%$quote");
				print F "  <tr><th>$field</th><td align=\"$align\">$vselect$selectopt*[end()]*</td></tr>\n";
			}
		} else {
			print F "  <tr><th>$field</th><td align=\"$align\">*[nl2br:$field]*</td></tr>\n";
		}
	}
	print F <<END;
</table>
*[end()]*
</BODY></HTML>
END
	close F;
}

sub addandcond {
	my($select, $condition) = @_;
	my $result = $select;
	$result =~ s/ where / where $condition and /i or
	$result =~ s/( order | group |$)/ where $condition$1/i;
	$result;
}

sub makeinserttemplate {
	my($dbh, $selfurl, $charset, $templatedir, $fileurl, $table, $number, 
		$types, $disps, $patterns, $marks, @fields) = @_;
	my $qtable = quotename($table);
	my $checkfuncstr = checkfuncstr($patterns, $marks, @fields);
	open F, ">$templatedir/table${number}i.html" or return;
	print "make template $templatedir/table${number}i.html\n";
	print F <<END;
<HTML><HEAD>
<meta http-equiv="Content-Type" content="text/html; charset=$charset">
<script language="javascript"><!--
$checkfuncstr
function dosubmit() {
	if( checkvalue() )
		document.form1.submit();
}
//--></script>
</HEAD>
<BODY BGCOLOR="white">
*[select(from $qtable)]*
<p align="right">
<a href="$selfurl?template=default.html">HOME</a>
<a href="$selfurl?template=table$number.html">$qtable table</a>
</p>
<h3>Insert new data into $qtable table</h3>
<form name="form1" method="POST" action="$selfurl" enctype="multipart/form-data" onSubmit="return checkvalue()">
<input type="hidden" name=".command" value="insert">
<input type="hidden" name=".fromwhere" value="*[urlescape:param(fromwhere)]*">
<input type="hidden" name=".msg_ok" value="inserted">
<input type="hidden" name=".url_ok" value="$selfurl?template=table$number.html">
<input type="hidden" name=".msg_error" value="cannot insert">
<input type="hidden" name=".url_error" value="$selfurl?template=table$number.html">
<table border=1>
END
	for my $field(@fields) {
		my $mark = $marks->{$field};
		my($cols, $rows, @cand, $mcand, $select, $selectval, $selectopt);
		my $disp = $disps->{$field};
		if( $disp =~ /^\s*\[(.+)\](m?)\s*$/ ) {
			@cand = split /\s*,\s*/, $1, -1;
			$mcand = $2;
		} elsif( $disp =~ /^\s*(\*\[select.+\]\*),([^,]+),([^,]+)(,m)?\s*$/ ) {
			($select, $selectval, $selectopt, $mcand) = ($1,$2,$3,$4);
			$selectopt = "*[$selectopt]*" unless $selectopt =~ /\*\[/;
		} elsif( $disp =~ /^\s*(\d+)(?:,(\d+))?/ ) {
			($cols, $rows) = ($1, $2);
		}
		if( $mark =~ /(@|##)/ ) {
			print F "  <tr><th>$field</th><td>(auto)<input type=\"hidden\" name=\"$mark$field\" value=\"\"></td></tr>\n";
		} elsif( $mark =~ /&/ ) {
			print F "  <tr><th>$field</th><td><input type=\"file\" name=\"$mark$field\" size=\"$cols\"></td></tr>\n";
		} elsif( @cand ) {
			if( $mcand ) {
				print F "  <tr><th>$field</th><td>";
				for my $cand(@cand) {
					print F "<input name=\"$mark$field\" type=\"checkbox\" value=\"$cand\">$cand ";
				}
				print F "</td></tr>\n";
			} else {
				print F "  <tr><th>$field</th><td><select name=\"$mark$field\">";
				for my $cand(@cand) {
					print F "<option>$cand</option>";
				}
				print F "</select></td></tr>\n";
			}
		} elsif( $select ) {
			if( $mcand ) {
				print F "  <tr><th>$field</th><td>$select*[repeat()]*\n<input name=\"$mark$field\" type=\"checkbox\" value=\"*[$selectval]*\">$selectopt \n*[end()]**[end()]*</td></tr>\n";
			} else {
				print F "  <tr><th>$field</th><td><select name=\"$mark$field\">$select*[repeat()]*\n<option value=\"*[$selectval]*\">$selectopt</option>\n*[end()]**[end()]*</select></td></tr>\n";
			}
		} elsif( $rows ) {
			print F "  <tr><th>$field</th><td><textarea name=\"$mark$field\" cols=\"$cols\" rows=\"$rows\"></textarea></td></tr>\n";
		} else {
			print F "  <tr><th>$field</th><td><input type=\"text\" name=\"$mark$field\" size=\"$cols\"></td></tr>\n";
		}
	}
	print F <<END;
</table>
<input type="button" value="submit" onClick="dosubmit()">
<input type="reset" value="reset">
</form>
*[end()]*
</BODY></HTML>
END
	close F;
}

sub makeedittemplate {
	my($dbh, $selfurl, $charset, $templatedir, $fileurl, $table, $number, 
		$primary, $types, $disps, $patterns, $marks, @fields) = @_;
	my $qtable = quotename($table);
	my $checkfuncstr = checkfuncstr($patterns, $marks, @fields);
	open F, ">$templatedir/table${number}e.html" or return;
	print "make template $templatedir/table${number}e.html\n";
	print F <<END;
<HTML><HEAD>
<meta http-equiv="Content-Type" content="text/html; charset=$charset">
<script language="javascript"><!--
$checkfuncstr
function dosubmit() {
	if( checkvalue() )
		document.form1.submit();
}
function deleteconfirm() {
	if( confirm("delete this data OK?") ) {
		return true;
	}
	return false;
}
//--></script>
</HEAD>
<BODY BGCOLOR="white">
*[select(from $qtable)]*
<p align="right">
<a href="$selfurl?template=default.html">HOME</a>
<a href="$selfurl?template=table$number.html">$qtable table</a>
<a href="$selfurl?template=table${number}v.html&where=*[urlescape:param(where)]*">view this data</a>
</p>
<h3>Edit data in $qtable table</h3>
<form name="form1" method="POST" action="$selfurl" enctype="multipart/form-data" onSubmit="return checkvalue()">
<input type="hidden" name=".command" value="update">
<input type="hidden" name=".fromwhere" value="*[urlescape:param(fromwhere)]*">
<input type="hidden" name=".msg_ok" value="updated">
<input type="hidden" name=".url_ok" value="$selfurl?template=table${number}e.html&where=*[urlescape:param(where)]*">
<input type="hidden" name=".msg_error" value="cannot update">
<input type="hidden" name=".url_error" value="$selfurl?template=table${number}e.html&where=*[urlescape:param(where)]*">
<table border=1>
END
	for my $field(@fields) {
		my $needquote = $types->{$field} =~ /char|text|date|time/i;
		my $mark = $marks->{$field};
		$mark = '#' if $mark eq '##';
		my($cols, $rows, @cand, $mcand, $select, $selectval, $selectopt);
		my $disp = $disps->{$field};
		if( $disp =~ /^\s*\[(.+)\](m?)\s*$/ ) {
			@cand = split /\s*,\s*/, $1, -1;
			$mcand = $2;
		} elsif( $disp =~ /^\s*(\*\[select.+\]\*),([^,]+),([^,]+)(,m)?\s*$/ ) {
			($select, $selectval, $selectopt, $mcand) = ($1,$2,$3,$4);
			$selectopt = "*[$selectopt]*" unless $selectopt =~ /\*\[/;
		} elsif( $disp =~ /^\s*(\d+)(?:,(\d+))?/ ) {
			($cols, $rows) = ($1, $2);
		}
		if( $field eq $primary ) {
			print F "  <tr><th>$field</th><td>*[$field]*<input type=\"hidden\" name=\"$mark$field\" value=\"*[$field]*\"></td></tr>\n";
		} elsif( $mark =~ /\@\@/ ) {
			print F "  <tr><th>$field</th><td>*[$field]*</td></tr>\n";
		} elsif( $mark =~ /\@/ ) {
			print F "  <tr><th>$field</th><td>*[$field]*<input type=\"hidden\" name=\"$mark$field\" value=\"\"></td></tr>\n";
		} elsif( $mark =~ /&/ ) {
			print F "  <tr><th>$field</th><td><a href=\"$fileurl/*[urlescape:$field]*\" target=\"_blank\">*[$field]*</a> <input type=\"file\" name=\"$mark$field\" size=\"$cols\"><input type=\"checkbox\" name=\".clear$mark$field\" value=\"1\">clear</td></tr>\n";
		} elsif( @cand ) {
			if( $mcand ) {
				print F "  <tr><th>$field</th><td>";
				if( $types->{$field} =~ /^listtable\((.+)\)/ ) {
					my($ltname, $ltf1name, $ltf1type, $ltf2name, $ltf2type, 
						$ltselect, $ltdisp) = splitlisttablespec($1);
					my $quote1 = $ltf1type =~ /char|text|date|time/i ? "'" : "";
					my $quote2 = $ltf2type =~ /char|text|date|time/i ? "'" : "";
					for my $cand(@cand) {
						print F "<input name=\"$mark$field\" type=\"checkbox\" value=\"$cand\" *[select(from \"$ltname\" where \"$ltf1name\" = $quote1%[$primary]%$quote1 and \"$ltf2name\" = $quote2$cand$quote2)]*checked*[else()]**[end()]*>$cand ";
					}
				} else {
					for my $cand(@cand) {
						print F "<input name=\"$mark$field\" type=\"checkbox\" value=\"$cand\" *[if('%[$field]%' =~ /\\Q$cand\\E/)]*checked*[end()]*>$cand ";
					}
				}
				print F "</td></tr>\n";
			} else {
				print F "  <tr><th>$field</th><td><select name=\"$mark$field\">";
				for my $cand(@cand) {
					print F "<option *[if('$cand' eq '%[$field]%')]*selected*[end()]*>$cand</option>";
				}
				print F "</select></td></tr>\n";
			}
		} elsif( $select ) {
			if( $mcand ) {
				if( $types->{$field} =~ /^listtable\((.+)\)/ ) {
					my($ltname, $ltf1name, $ltf1type, $ltf2name, $ltf2type, 
						$ltselect, $ltdisp) = splitlisttablespec($1);
					my $quote1 = $ltf1type =~ /char|text|date|time/i ? "'" : "";
					my $quote2 = $ltf2type =~ /char|text|date|time/i ? "'" : "";
					print F "  <tr><th>$field</th><td>*[embed(<<END,pv,%[$primary]%)]*$select*[repeat()]*\n<input name=\"$mark$field\" type=\"checkbox\" value=\"*[$selectval]*\" *[select(from \"$ltname\" where \"$ltf1name\" = $quote1%[param(pv)]%$quote1 and \"$ltf2name\" = $quote2%[$selectval]%$quote2)]*checked*[else()]**[end()]*>$selectopt \n*[end()]**[end()]*\nEND\n</td></tr>\n";
				} else {
					print F "  <tr><th>$field</th><td>*[embed(<<END,fv,%[$field]%)]*$select*[repeat()]*\n<input name=\"$mark$field\" type=\"checkbox\" value=\"*[$selectval]*\" *[if('%[param(fv)]%' =~ /\\Q%[$selectval]%\\E/)]*checked*[end()]*>$selectopt \n*[end()]**[end()]*\nEND\n</td></tr>\n";
				}
			} else {
				print F "  <tr><th>$field</th><td><select name=\"$mark$field\">*[embed(<<END,fv,%[$field]%)]*$select*[repeat()]*\n<option value=\"*[$selectval]*\" *[if('%[$selectval]%' eq '%[param(fv)]%')]*selected*[end()]*>$selectopt</option>\n*[end()]**[end()]*\nEND\n</select></td></tr>\n";
			}
		} elsif( $rows ) {
			print F "  <tr><th>$field</th><td><textarea name=\"$mark$field\" cols=\"$cols\" rows=\"$rows\">*[$field]*</textarea></td></tr>\n";
		} else {
			print F "  <tr><th>$field</th><td><input type=\"text\" name=\"$mark$field\" size=\"$cols\" value=\"*[$field]*\"></td></tr>\n";
		}
	}
	print F <<END;
</table>
<input type="button" value="submit" onClick="dosubmit()">
<input type="reset" value="reset">
</form>
<form method="POST" action="$selfurl" onSubmit="return deleteconfirm()">
<input type="hidden" name=".command" value="delete">
<input type="hidden" name=".fromwhere" value="*[urlescape:param(fromwhere)]*">
<input type="hidden" name=".msg_ok" value="deleted">
<input type="hidden" name=".url_ok" value="$selfurl?template=table${number}.html">
<input type="hidden" name=".msg_error" value="cannot delete">
<input type="hidden" name=".url_error" value="$selfurl?template=table${number}.html">
<input type="submit" value="delete this data">
</form>
*[end()]*
</BODY></HTML>
END
	close F;
}

sub checkfuncstr {
	my($patterns, $marks, @fields) = @_;
	my $checkfuncstr = "function checkvalue() { var errmsg = ''; \n";
	for my $field(@fields) {
		my $mark = $marks->{$field};
		$mark = '#' if $mark eq '##';
		my $pattern = $patterns->{$field};
		if( $pattern =~ /\.\"(.+?)\"/ ) {
			my $msg = $1;
			$checkfuncstr .= qq#if( document.form1['$mark$field'].value == "" ) { errmsg = errmsg + "$msg\\n"; }\n#;
		}
		if( $pattern =~ /\/(.+)\/\"(.+?)\"/ ) {
			my($ptn, $msg) = ($1, $2);
			$checkfuncstr .= qq#if( document.form1['$mark$field'].value != "" && !document.form1['$mark$field'].value.match(/$ptn/) ) { errmsg = errmsg + "$msg\\n"; }\n#;
		}
	}
	$checkfuncstr .= qq[if( errmsg != "" ) { alert(errmsg); return false; }\nreturn true;\n}];
	$checkfuncstr;
}

sub do_csv {
	my($cgi, $cfg, $para) = @_;
	my $csvtable = $cfg->{'.csvtable'}
		or fatal($cgi, $cfg->{charset}, 
		'csv command needs .csvtable parameter');
	my $csvfile = $cfg->{'.csvfile'}
		or fatal($cgi, $cfg->{charset}, 
		'csv command needs .csvfile parameter');
	my $fieldmarks = fieldmarks($cfg, $csvtable);
	fatal($cgi, $cfg->{charset}, "missing \"$csvtable\" table")
		unless $fieldmarks;
	my $table = $csvtable;
	$csvtable = quotename($csvtable);
	my($dbh) = do_prepare('csv', $cgi, $cfg);
	my $result = "$csvtable<br>\n";
	if( $cfg->{'.cleartable'} ) {
		$result .= "clear table<br>\n";
		my $sql = "delete from $csvtable";
		my $rv = $dbh->do($sql) + 0;
	}
	my $linenum = 0;
	my(@fields, $fields, @isnum, @autotimefields, @autonumfields);
	while(<$csvfile>) {
		$linenum++;
		my @array = csv2array($_);
		while( ref($array[0]) ) {
			@array = csv2array(scalar(<$csvfile>), @array);
		}
		#print "[",join('|',@array),"]\n";
		if( $linenum == 1 ) {
			@fields = @array;
			my @unknownfileds;
			for( my $j = 0; $j < @fields; $j++ ) {
				my $field = $fields[$j];
				if( exists $fieldmarks->{$field} ) {
					if( $fieldmarks->{$field} =~ /#/ ) {
					#if( $fields[$j] =~ s/^#// ) {
						$isnum[$j] = 1;
					}
					$fieldmarks->{$field} = '';
				} else {
					push @unknownfileds, $field;
				}
			}
			if( @unknownfileds ) {
				$result .= "$linenum: ERROR! unknown fields: ".
					join(" ",@unknownfileds) ."<br>\n";
				last;
			}
			for my $field(keys %$fieldmarks) {
				if( $fieldmarks->{$field} eq '##' ) {
					push @autonumfields, $field;
				} elsif( $fieldmarks->{$field} eq '@' ) {
					push @autotimefields, $field;
				}
			}
			$fields = join(',', map {quotename($_)} 
				(@fields, @autonumfields, @autotimefields));
			$result .= "$linenum: $fields<br>\n";
		} elsif( @array == @fields ) {
			for( my $j = 0; $j < @array; $j++ ) {
				if( $isnum[$j] ) {
					$array[$j] = 'null' if $array[$j] eq '';
				} else {
					$array[$j] = $dbh->quote($array[$j]) 
				}
			}
			for(@autonumfields) {
				push @array, "coalesce((select max(".quotename($_).") from $csvtable), 0) + 1";
			}
			for(@autotimefields) {
				push @array, "'".now()."'";
			}
			my $values = join(',', @array);
			my $sql = "insert into $csvtable ($fields) values ($values)";
			my $rv = $dbh->do($sql) + 0;
			$result .= $rv ? "$linenum: $values<br>\n" :
				"$linenum: $values: ERROR!<br>\n";
			if( !$rv && $cfg->{verbose} ) {
				$result .= " : $sql : ".$dbh->errstr."<br>\n";
			}
		} else {
			$result .= "$linenum: ERROR! number or fields mismatch<br>\n";
		}
	}
	my $selfurl = $cgi->url;
	my $tablenum = $para->{tablenum};
	result($cgi, $cfg->{charset}, 'csv result', $result, "$selfurl?template=table$tablenum.html");
}

sub fields {
	my($cfg, $table) = @_;
	my $fh = new FileHandle "$cfg->{templatedir}/tablefields";
	my @tf;
	while(<$fh>) {
		chomp;
		if( /^\Q$table\E\t/ ) {
			(undef, @tf) = split /\t/;
			last;
		}
	}
	close $fh;
	return @tf;
}

sub fieldmarks {
	my($cfg, $table) = @_;
	my @tf = fields($cfg, $table);
	if( @tf ) {
		my %fieldmarks;
		for(@tf) {
			if( s/^([*@#&]+)// ) {
				$fieldmarks{$_} = $1;
			} else {
				$fieldmarks{$_} = '';
			}
		}
		return \%fieldmarks;
	} else {
		return;
	}
}

sub do_select {
	my($cgi, $cfg, $fromwhere, $para) = @_;
	my($dbh, $table, $where) = do_prepare('select', $cgi, $cfg, $fromwhere);
	
	my $tfile = $cfg->{'.template'};
	unless( $tfile ) {
		fatal($cgi, $cfg->{charset}, "select needs .template parameter");
	}
	my $templateclass = $cfg->{templateclass} || 'Greese::Template::Html';
	my $tobj = $templateclass->new("$cfg->{templatedir}/$tfile");
	my $robj = Greese::Report->new($tobj);
	
	my $andor = $cfg->{'.andor'} || 'and';
	$where = "($where)" if $where;
	for my $field(keys %$para) {
		my $value = $para->{$field};
		next if $value eq '';
		my $isnum = $field =~ s/^#//;
		my $op = $cfg->{".op_$field"} || '=';
		if( $op =~ /^like$/i && $value !~ /%/ ) {
			$value = '%'.$value.'%';
		}
		$value = $dbh->quote($value) unless $isnum;
		$field = quotename($field);
		$where .= " $andor " if $where;
		$where .= "$field $op $value";
	}
	
	print $cgi->header(-type => 'text/html', -charset => $cfg->{charset});
	my %param;
	$param{dbihandle} = $dbh;
	$param{cgi} = $cgi->url;
	$param{empty} = 1;
	$param{where} = $where;
	$robj->print(%param);
}

sub csv2array {
	my($line) = shift;
	#$line = Jconv($line, 'EUC', 'SJIS');
	my $lastref = shift;
	$line = $$lastref . $line if $lastref;
	my @result;
	local($_) = $line;
	s/[\r\n]+$//;
	while( !/^$/ && (/^ *(\".*?\") *(,|$)/s || /^ *([^,]*) *(,|$)/ ) ) { 
		push(@result,$1); 
		$_ = $'; 
		push(@result,'') if $2 eq ',' && $_ eq '';
	}
	if( $result[$#result] =~ /^\"/ && $' !~ /\"$/ ) {
		return \$line;
	} else {
		map { s/^\"// && s/\"$// && s/\"\"/\"/g } @result;	# for Excel CSV
		return @result;
	}
}

sub do_prepare {
	my($method, $cgi, $cfg, $fromwhere) = @_;
	if( $cfg->{"${method}_password"} &&
		$cfg->{'.password'} ne $cfg->{"${method}_password"} ) {
		result($cgi, $cfg->{charset}, 'authorization result', 
			'password mismatch', 'back()');
		exit;
	}
	if( $cfg->{"${method}_allowuser"} ) {
		my $user = $cgi->user_name || $cfg->{authuser};
		unless( grep {$_ eq $user} 
			split(/\s*,\s*/, $cfg->{"${method}_allowuser"}) ) {
			result($cgi, $cfg->{charset}, 'authorization result', 
				"user '$user' is not allowed to execution", 'back()');
			exit;
		}
	}
	if( $cfg->{"${method}_allowgroup"} ) {
		my $group = $cfg->{authgroup};
		unless( grep {$_ eq $group} 
			split(/\s*,\s*/, $cfg->{"${method}_allowgroup"}) ) {
			result($cgi, $cfg->{charset}, 'authorization result', 
				"group '$group' is not allowed to execution", 'back()');
			exit;
		}
	}
	my $dbh = DBI->connect($cfg->{db}, $cfg->{user}, $cfg->{password}, 
		{AutoCommit => 1});
	fatal($cgi, $cfg->{charset}, "cannot connect to the database '$cfg->{db}'") 
		unless $dbh;
	my $checkwhere;
	if( $cfg->{'.password_field'} ) {
		$checkwhere = quotename($cfg->{'.password_field'})." = ".
			$dbh->quote($cfg->{'.password'});
	}
	my($tables) = $fromwhere =~ /from (.+?)(where|order|group|$)/i;
	my($where) = $fromwhere =~ /( where .+?)(order|group|$)/i;
	if( $checkwhere ) {
		$where = $where ? "$where and " : "where ";
		$where .= $checkwhere;
	}
	my @tables = split /\s*,\s*/, $tables;
	fatal($cgi, $cfg->{charset}, 'cannot update multiple tables') 
		if @tables > 1;
	$dbh->trace($cfg->{dbitrace}) if $cfg->{dbitrace};
	($dbh, $tables, $where);
}

sub do_exec {
	my($cgi, $dbh, $cfg, $where, @sqls) = @_;
	$where =~ s/^\s*where\s+//;
	my($msg, $vmsg, $url);
	my $rv = 1;
	for my $sql(@sqls) {
		my $r = $dbh->do($sql) + 0;
		$rv = 0 unless $r;
		if( $cfg->{verbose} ) {
			$vmsg .= "<br> : $sql : ".$dbh->errstr;
			$vmsg .= '<br>updated records: '.($r == -1 ? 'unknown' : $r);
		}
	}
	if( $rv ) {
		if( $cfg->{'.template_ok'} ) {
			templateresult($cgi, $cfg, $dbh, $cfg->{'.template_ok'}, where => $where);
		} else {
			($msg, $url) = ($cfg->{'.msg_ok'}, $cfg->{'.url_ok'});
			$msg .= $vmsg;
			result($cgi, $cfg->{charset}, 'result', $msg, $url);
		}
	} else {
		if( $cfg->{'.template_error'} ) {
			templateresult($cgi, $cfg, $dbh, $cfg->{'.template_error'}, 
				where => $where);
		} else {
			($msg, $url) = ($cfg->{'.msg_error'}, $cfg->{'.url_error'});
			$msg .= $vmsg;
			result($cgi, $cfg->{charset}, 'result', $msg, $url);
		}
	}
	$rv;
}


# obsolete
sub readcfgfile {
	my($cfgfile, %cfg) = @_;
	my($invalue, $endmark, $key, $value);
	my $F = new FileHandle;
	open($F, "$cfgfile");
	$invalue = 0;
	while(<$F>) {
		if( $invalue ) {
			if( /^$endmark$/ ) { 
				$invalue = 0; 
			} else { 
				$cfg{$key} .= $_; 
			}
		} elsif( /^(\w+)\s*=\s*(.+)/ ) {
			$key = $1;
			$value = $2;
			if( $value =~ /^<<(.+)/ ) {
				$endmark = $1;
				$invalue = 1;
			} else { 
				$value =~ s/^\"// and $value =~ s/\"\s*$//;
				$cfg{$key} = $value; 
			}
		}
	}
	close($F);
	return %cfg;
}

sub url_escape {
	my($str) = @_;
	$str =~ s/([^\/:+\$,A-Za-z0-9\-_.!~*'()])/sprintf("%%%02X",ord($1))/ge;
	$str;
}

sub url_unescape {
	my($str) = @_;
	$str =~ s/%([0-9A-Fa-f]{2})/chr(hex($1))/eg;
	$str;
}

sub quotename {
	my($name) = @_;
	$name =~ s/[^\s\.]+/\"$&\"/g;
	$name;
}

sub dequotename {
	my($name) = @_;
	$name =~ s/\"([^\s\.]+)\"/$1/g;
	$name =~ s/^\s+//;
	$name =~ s/\s+$//;
	$name;
}

sub escape_quote {
	my($str) = @_;
	$str =~ s/\"/\"\"/g;
	$str;
}

sub now {
	my($s,$m,$h,$D,$M,$Y) = localtime;
	$Y += 1900;
	$M++;
	sprintf "%04d-%02d-%02d %02d:%02d:%02d",$Y,$M,$D,$h,$m,$s;
}

1;
__END__
=pod

=head1 NAME

GreeseCGI - CGI with Greese

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 AUTHOR

=head1 SEE ALSO

=cut
